import subprocess
from pathlib import Path

def get_latest_commit_for_file(filename: str) -> str:
    try:
        git_hash = subprocess.check_output(
            ["git", "log", "-n", "1", "--pretty=format:%H", "--", filename],
            text=True
        ).strip()
        if not git_hash:
            raise ValueError(f"No commit history for {filename}")
        return git_hash
    except Exception as e:
        if isinstance(e, ValueError) and "No commit history" in str(e):
            raise
        raise ValueError(f"Error retrieving git history for {filename}")

def is_file_tracked_by_git(filename: str) -> bool:
    try:
        result = subprocess.run(
            ["git", "ls-files", "--error-unmatch", filename],
            capture_output=True, text=True, check=True
        )
    except subprocess.CalledProcessError:
        return False
    return True

def is_git_hash_on_file(filename: str, git_hash: str) -> bool:
    if not git_hash: return True

    change_list = subprocess.run(
        ["git", "log", "--pretty=format:%H", "--", filename],
        capture_output=True,
        text=True,
        check=True
    ).stdout.splitlines()

    return git_hash in change_list
    

def is_ancestor(commit1: str, commit2: str) -> bool:
    '''
    Given an 2 commits, return True if commit1 is an ancestor of commit2 else False
    '''
    try:
        result = subprocess.run(
            ["git", "merge-base", "--is-ancestor", commit1, commit2],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=False
        )
        return result.returncode == 0
    except subprocess.CalledProcessError:
        print(f"Error checking commit ancestry for {commit1} {commit2}")
        return False




def get_closest_ancestor_hash(filename: str, git_hash: str) -> str:
    '''
    Given a filename and git_hash, return the hash that last changed the file which is right before the provided hash
    '''
    print(f"Locating closest commit for {filename} that is right before the commit {git_hash}")
    # get a list of git-hashes that changed the supplied file
    change_list = subprocess.run(
        ["git", "log", "--pretty=format:%H", "--", filename],
        capture_output=True,
        text=True,
        check=True
    ).stdout.splitlines()
    
    # loop through the hashes
    for i in range(len(change_list)):
        # if current hash is before the specified git_hash and the previous one wasn't, then return it
        if is_ancestor(change_list[i], git_hash):
            return change_list[i]
    raise ValueError("Provided git_hash is invalid")


import pytest 
import subprocess
from scimon import utils

def test_get_latest_commit_for_file_success(monkeypatch):
    mock_file = "foo.txt"
    mock_git_commit = "abcd1234"
    monkeypatch.setattr(
        utils.subprocess, "check_output",
        lambda *args, **kwargs: mock_git_commit
    )
    assert utils.get_latest_commit_for_file(mock_file) == mock_git_commit

def test_get_latest_commit_for_file_no_history(monkeypatch):
    mock_file = "foo.txt"
    monkeypatch.setattr(
        utils.subprocess, "check_output",
        lambda *args, **kwargs: ""
    )
    with pytest.raises(ValueError) as ei:
        utils.get_latest_commit_for_file(mock_file)
    assert f"No commit history for {mock_file}" in str(ei.value)

def test_get_latest_commit_for_file_git_throwing_error(monkeypatch):
    def throw(*args, **kwargs):
        raise Exception("git failed")
    
    mock_file = "foo.txt"
    monkeypatch.setattr(utils.subprocess, "check_output", throw)
    with pytest.raises(ValueError) as ei:
        utils.get_latest_commit_for_file(mock_file)
    assert f"Error retrieving git history for {mock_file}" in str(ei.value)

def test_is_file_tracked_by_git_true(monkeypatch):
    def true(cmd, capture_output, text, check):
        return subprocess.CompletedProcess(cmd, 0, stdout='', stderr='')
    
    mock_file = "foo.txt"
    monkeypatch.setattr(utils.subprocess, "run", true)
    assert utils.is_file_tracked_by_git(mock_file) is True

def test_is_file_tracked_by_git_false(monkeypatch):
    def throw(cmd, capture_output, text, check):
        raise subprocess.CalledProcessError(1, cmd)
    
    mock_file = "foo.txt"
    monkeypatch.setattr(utils.subprocess, "run", throw)
    assert utils.is_file_tracked_by_git(mock_file) is False

def test_is_git_hash_on_file_empty_hash(monkeypatch):
    assert utils.is_git_hash_on_file("foo.txt", "") is True

def test_is_git_hash_on_file_true(monkeypatch):
    mock_hash = "abcd1234"

    def git_log(cmd, capture_output, text, check):
        return subprocess.CompletedProcess(cmd, 0, stdout=mock_hash, stderr='')

    monkeypatch.setattr(utils.subprocess, "run", git_log)
    assert utils.is_git_hash_on_file("foo.txt", mock_hash) is True

def test_is_git_hash_on_file_false(monkeypatch):
    mock_hash = "abcd1234"

    def git_log(cmd, capture_output, text, check):
        return subprocess.CompletedProcess(cmd, 0, stdout=mock_hash, stderr='')

    monkeypatch.setattr(utils.subprocess, "run", git_log)
    assert utils.is_git_hash_on_file("foo.txt", "heheheha") is False

def test_is_ancestor_true(monkeypatch):
    mock_commit_1 = "abcd1234"
    mock_commit_2 = "defg5678"

    def git_merge_base(cmd, stdout, stderr, check):
        return subprocess.CompletedProcess(cmd, 0)
    
    monkeypatch.setattr(utils.subprocess, "run", git_merge_base)
    assert utils.is_ancestor(mock_commit_1, mock_commit_2) is True

def test_is_ancestor_false(monkeypatch):
    mock_commit_1 = "abcd1234"
    mock_commit_2 = "defg5678"

    def git_merge_base(cmd, stdout, stderr, check):
        return subprocess.CompletedProcess(cmd, 1)
    
    monkeypatch.setattr(utils.subprocess, "run", git_merge_base)
    assert utils.is_ancestor(mock_commit_1, mock_commit_2) is False

def test_is_ancestor_error(monkeypatch):
    mock_commit_1 = "abcd1234"
    mock_commit_2 = "defg5678"

    def throw(*args, **kwargs):
        raise subprocess.CalledProcessError(2, args[0])
    
    monkeypatch.setattr(utils.subprocess, "run", throw)

    assert utils.is_ancestor(mock_commit_1, mock_commit_2) is False
    
def test_get_closest_ancestor_hash_found(monkeypatch):
    mock_ancestor_hashes="abc\ndef\nghi\n"
    def git_log(cmd, capture_output, text, check):
        return subprocess.CompletedProcess(cmd, 0, stdout=mock_ancestor_hashes)
    
    monkeypatch.setattr(utils.subprocess, "run", git_log)
    monkeypatch.setattr(utils, "is_ancestor", lambda c1, c2: c1 == "def")
    
    assert utils.get_closest_ancestor_hash("foo.txt", "xyz") == "def"


def test_get_closest_ancestor_hash_invalid(monkeypatch):
    mock_ancestor_hashes="abc\ndef\nghi\n"
    def git_log(cmd, capture_output, text, check):
        return subprocess.CompletedProcess(cmd, 0, stdout=mock_ancestor_hashes)
    
    monkeypatch.setattr(utils.subprocess, "run", git_log)
    monkeypatch.setattr(utils, "is_ancestor", lambda c1, c2: False)
    
    with pytest.raises(ValueError) as ei:
        utils.get_closest_ancestor_hash("foo.txt", "xyz")
    
    assert "Provided git_hash is invalid" in str(ei.value)





import sqlite3
from scimon.models import Process, File, Edge
from typing import List, Tuple
DB_NAME=".db"


def get_db() -> sqlite3.Connection:
    con = sqlite3.connect(DB_NAME)
    print("Database connection acquired")
    return con

def get_processes_trace(commit_hash: str, db: sqlite3.Connection) -> List[Tuple]:
    '''Returns a list of (parent_pid, pid, child_pid, syscall) for a given commit hash'''
    cursor = db.cursor()
    processes_sql = '''SELECT DISTINCT parent_pid, pid, child_pid, syscall FROM processes WHERE commit_hash = ?'''
    cursor.execute(processes_sql, (commit_hash,))
    return cursor.fetchall()


def get_opened_files_trace(commit_hash: str, db: sqlite3.Connection) -> List[Tuple]:
    '''Returns a list of (pid, filename, syscall, mode, open_flag) for a given commit hash'''
    cursor = db.cursor()
    opened_files_sql = '''SELECT DISTINCT pid, filename, syscall, mode, open_flag FROM opened_files WHERE commit_hash = ?'''
    cursor.execute(opened_files_sql, (commit_hash,))
    return cursor.fetchall()

def get_executed_files_trace(commit_hash: str, db: sqlite3.Connection) -> List[Tuple]:
    '''Returns a list of (pid, filename, syscall) for a given commit hash'''
    cursor = db.cursor()
    executed_files_sql = '''SELECT pid, filename, syscall FROM executed_files WHERE commit_hash = ?'''
    cursor.execute(executed_files_sql, (commit_hash,))
    return cursor.fetchall()

def get_command(commit_hash: str, db: sqlite3.Connection) -> str:
    '''Returns the command associated where commit_hash is the post command commit hash in the commands table'''
    cursor = db.cursor()
    get_command_sql = '''SELECT command FROM commands WHERE post_command_commit = ?'''
    cursor.execute(get_command_sql, (commit_hash,))
    return cursor.fetchall()[0][0]


from typing import Optional
from scimon.models import Graph, Node, Edge
from scimon.db import *
from scimon.utils import *
import os
from jinja2 import Template
from pathlib import Path

MAKE_FILE_RULE_TEMPLATE = Template("""
{{ target }}: {{ prerequisites }}
\t{{ recipe }}
""")

MAKE_FILE_NAME='reproduce.mk'

def get_trace_data(git_hash: str, db) -> Tuple[list, list, list]:
    """Retrieve all trace data for a given git hash."""
    print("Getting trace data")
    processes_trace = get_processes_trace(git_hash, db)
    open_files_trace = get_opened_files_trace(git_hash, db)
    executed_files_trace = get_executed_files_trace(git_hash, db)
    return processes_trace, open_files_trace, executed_files_trace


def build_process_nodes_and_edges(graph: Graph, processes_trace: list, git_hash: str):
    """Build process nodes and their relationships in the graph."""
    print("Building process nodes and edges")

    for pt in processes_trace:
        parent_pid, pid, child_pid, syscall = pt

        parent_process_node = Process(git_hash=git_hash, pid=parent_pid)
        process_node = Process(git_hash=git_hash, pid=pid)
        child_process_node = Process(git_hash=git_hash, pid=child_pid)
        
        parent_edge = Edge(parent_process_node, process_node, syscall)
        child_edge = Edge(process_node, child_process_node, syscall)

        graph.add_node(parent_process_node)
        graph.add_node(process_node)
        graph.add_node(child_process_node)
        graph.add_edge(parent_edge)
        graph.add_edge(child_edge)


def build_file_read_write_nodes_and_edges(graph: Graph, file_traces: list, git_hash: str, is_execution: bool = False):
    """Build file nodes and their relationships to processes."""
    print("Building file read write nodes and edges")
    for trace in file_traces:
        pid, filename, syscall, mode, open_flag = trace
        # filter files not part of the git repository
        if not is_file_tracked_by_git(filename):
            continue
        if os.path.isdir(filename):
            continue
        # normalize filename
        cwd = Path(os.getcwd())
        abs_path = Path(filename).resolve()
        filename = str(abs_path.relative_to(cwd))

        file_node = File(git_hash, filename)
        # if file with same path already in the graph, fetch that node in the graph
        # TODO:
        process_node = Process(git_hash=git_hash, pid=pid)
        if "O_WRONLY" in open_flag or "O_CREAT" in open_flag or "O_RDWR" in open_flag or "O_TRUNC" in open_flag:
            process_to_file_edge = Edge(file_node, process_node, syscall)
        else:
            process_to_file_edge = Edge(process_node, file_node, syscall)
        graph.add_node(file_node)
        graph.add_node(process_node)
        graph.add_edge(process_to_file_edge)


def build_file_execution_nodes_and_edges(graph: Graph, file_traces: list, git_hash: str, is_execution: bool = False):
    """Build file nodes and their relationships to processes."""
    print("Building file execution nodes and edges")
    for trace in file_traces:
        pid, filename, syscall = trace
        # filter files not part of the git repository
        if not is_file_tracked_by_git(filename):
            continue
        if os.path.isdir(filename):
            continue
        # normalize filename
        cwd = Path(os.getcwd())
        abs_path = Path(filename).resolve()
        filename = str(abs_path.relative_to(cwd))


        file_node = File(git_hash, filename)
        
        process_node = Process(git_hash=git_hash, pid=pid)
        process_to_file_edge = Edge(process_node, file_node, syscall)

        graph.add_node(file_node)
        graph.add_node(process_node)
        graph.add_edge(process_to_file_edge)


def generate_graph(filename: str, git_hash: str) -> Graph:
    '''
    Produce a provenance graph for a given file at a version of the given githash, 
    '''
    # Initialize
    graph = Graph()
    db = get_db()
    
    print(f"Preparing to generate graph for file {filename} with version {git_hash}")

    processes_trace, open_files_trace, executed_files_trace = get_trace_data(git_hash, db)

    build_process_nodes_and_edges(graph, processes_trace, git_hash)
    build_file_read_write_nodes_and_edges(graph, open_files_trace, git_hash)
    build_file_execution_nodes_and_edges(graph, executed_files_trace, git_hash)

    return graph



def reproduce(file: str, git_hash: Optional[str]):

    # TODO: check if cwd is a valid directory being monitored
    cwd = Path(os.getcwd())

    # Normalize file
    abs_path = Path(file).resolve()
    file = str(abs_path.relative_to(cwd))

    # Check if the file is a directory
    if os.path.isdir(file):
        print(f"{file} is a directory, skipping...")
        return

    # Check if the file exists in the git repository
    if not is_file_tracked_by_git(filename=file):
        print(f"{file} is not being tracked by the git repository")
        return
    
    # Check if the git hash exists on this file change list
    if not is_git_hash_on_file(file, git_hash):
        print(f"The provided git commit hash {git_hash} does not have any changes related to the file {file}")
        return 
    if not git_hash: 
        git_hash = get_latest_commit_for_file(file)

    # generate a file dependency graph containing the current node
    graph = generate_graph(file, git_hash)
    # traverse up the graph to get parents
    adj = graph.get_adj_list()

    # for k,v in adj.items():
    #     if isinstance(k, File):
    #         print(k.filename, ":", end=" ")
    #     else:
    #         print(k.pid, ":", end=" ")
    #     for n in v:
    #         if isinstance(n, File):
    #             print(n.filename, end=" ")
    #         else:
    #             print(n.pid, end=" ")
    #     print()
    
    if File(git_hash, file) not in adj:
        print(f"The current file {file} has no dependencies, directly checking the version {git_hash} out from git...")
        rule = MAKE_FILE_RULE_TEMPLATE.render(target=file, prerequisites="", recipe=f"git restore --source={git_hash} -- {file}")
        with open(MAKE_FILE_NAME, 'a') as f:
            f.write(rule)
        return

    dependencies = set()

    def dfs(node: Node):
        if node in adj:
            for parent in adj[node]:
                if isinstance(parent, File):
                    if parent.filename not in dependencies:
                        print(f"Parent file {parent.filename} of {file} located, calling reproduce on it...")
                        target_hash = get_closest_ancestor_hash(parent.filename, git_hash)
                        # call reproduce on that version
                        dependencies.add(parent.filename)
                        reproduce(parent.filename, target_hash)
                else:
                    print(f"Process {parent.pid} located from traversing the provenance graph, continuing traversing")
                    dfs(parent)
        
    dfs(File(git_hash, file))
    print("Fetching command from database")
    command = get_command(git_hash, get_db())
    # create the make rule
    rule = MAKE_FILE_RULE_TEMPLATE.render(target=file, prerequisites=" ".join(dependencies), recipe=command)
    with open(MAKE_FILE_NAME, 'a') as f:
        f.write(rule)





# def main():
#     parser = argparse.ArgumentParser(description="A passive scientific reproducibility tool")
#     subparsers = parser.add_subparsers(dest="command")

#     # reproduce [file] --git-hash=...
#     reproduce_parser = subparsers.add_parser("reproduce", help="Reproduce a given file")
#     reproduce_parser.add_argument("file", nargs=1, help="The filename of the file you want to reproduce")
#     reproduce_parser.add_argument("--git-hash", nargs="?", help="Specific git version of the file")

#     # TODO
#     # add [directory]
#     # list [directory]
#     # remove [directory]
#     args = parser.parse_args()
#     if args.command == "reproduce":
#         reproduce(args.file[0], args.git_hash)

# if __name__ == "__main__":
#     main()


from typing import Optional
import typer
from scimon import __app_name__, __version__
from scimon.scimon import reproduce as r
import os
app = typer.Typer()

def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f"{__app_name__} v{__version__}")
        raise typer.Exit()

@app.callback()
def main(version: Optional[bool] = typer.Option(
    None,
    "--version",
    "-v",
    help="Display the application's version and exit",
    callback=_version_callback,
    is_eager=True
)) -> None:
    return

@app.command(help="Generates a Makefile for reproducing the supplied file at a given version specified with the git commit hash.")
def reproduce(
    file: str = typer.Argument(help="Path to the file to reproduce"),
    git_hash: Optional[str] = typer.Option(None, "--git-hash", "-g", help="Git commit hash of the version to reproduce, selects newest version by default")
) -> None:
    r(file, git_hash)

@app.command(help="Initialize the current working directory for monitoring")
def init() -> None:
    # Adds cwd to ~/.scimon/.autogitcheck
    # TODO
    pass

@app.command(help="Lists all directories currently being monitored")
def list() -> None:
    # TODO
    pass

@app.command(help="Removes a directory from being monitored")
def remove(dir: str = typer.Argument(help="Directory to remove", default=os.getcwd())) -> None:
    # TODO
    pass


from typing import Optional, Set, Dict
import graphviz

class Node(object):
    def __init__(self, git_hash: str):
        self.git_hash = git_hash

class Process(Node):
    def __init__(self, git_hash: str, pid: int):
        super().__init__(git_hash)
        self.pid = pid
    def __eq__(self, other):
        return isinstance(other, Process) and self.git_hash == other.git_hash and self.pid == other.pid
    
    def __hash__(self):
        return hash((self.git_hash, self.pid))

class File(Node):
    def __init__(self, git_hash: str, filename: str):
        super().__init__(git_hash)
        self.filename = filename

    def __eq__(self, other):
        return isinstance(other, File) and self.git_hash == other.git_hash and self.filename == other.filename
    
    def __hash__(self):
        return hash((self.git_hash, self.filename))
    
class Edge(object):
    def __init__(self, in_node: Node, out_node: Node, syscall: str):
        self.in_node = in_node
        self.out_node = out_node
        self.syscall = syscall
    
    def __eq__(self, other):
        return isinstance(other, Edge) and self.in_node == other.in_node and self.out_node == other.out_node and self.syscall == other.syscall
    
    def __hash__(self):
        return hash((self.in_node, self.out_node, self.syscall))

class Graph(object):

    def __init__(self, nodes: Optional[Set[Node]] = None, edges: Optional[Set[Edge]] = None):
        self.nodes = set() if not nodes else nodes
        self.edges = set() if not edges else edges
        

    def add_node(self, node: Node) -> None:
        '''Adds the provided node into the collection of nodes in the graph, if it already exists then perform update if possible'''
        if node not in self.nodes:
            self.nodes.add(node)
    
    def add_edge(self, edge: Edge) -> None:
        '''
        Adds the provided edge into the collection of edges in the graph
        If one of in-node or out-node is not part of the graph, we add it into the graph as well
        '''
        
        self.add_node(edge.in_node)
        
        self.add_node(edge.out_node)
        
        self.edges.add(edge)


    def generate_dot(self, output_name="prov"):
        '''
        Generates a DOT graph visualization.
        '''
        dot = graphviz.Digraph(
            format='png', 
            graph_attr={
                'rankdir': 'TB',      
                'concentrate': 'true',
                'nodesep': '0.3',     
                'ranksep': '0.5',   
                'splines': 'line',  
                'overlap': 'false',   
                'fontname': 'Arial' 
            },
            node_attr={
                'shape': 'box',
                'style': 'rounded,filled',
                'fontname': 'Arial',
                'margin': '0.2,0.1'
            },
            edge_attr={
                'fontname': 'Arial',
                'fontsize': '10'
            }
        )

        # Process nodes with distinct styling
        for n in self.nodes:
            if isinstance(n, Process):
                attrs = {
                    'label': f"Process {n.pid}",
                    'git_hash': str(n.git_hash),
                    'fillcolor': "#C7E0F8",  # Light blue
                    'tooltip': f"PID: {n.pid}, Git: {n.git_hash}"
                }
                print(f"Node Added: {n.pid, n.git_hash}")
                dot.node(str(n.pid), **attrs)
            elif isinstance(n, File):
                attrs = {
                    'label': n.filename,
                    'git_hash': str(n.git_hash),
                    'fillcolor': "#F4F8CB",  # Light yellow
                    'tooltip': f"File: {n.filename}, Git: {n.git_hash}"
                }
                print(f"Node Added: {n.filename, n.git_hash}")
                dot.node(n.filename, **attrs)

        # Add edges with improved styling
        for e in self.edges:
            in_node_name = str(e.in_node.pid) if isinstance(e.in_node, Process) else e.in_node.filename
            out_node_name = str(e.out_node.pid) if isinstance(e.out_node, Process) else e.out_node.filename
            print(f"Edge Added: {in_node_name, out_node_name, e.syscall}")
            dot.edge(in_node_name, out_node_name, label=e.syscall)
            
        # Use dot layout engine for hierarchical layout
        dot.engine = 'dot'
        dot.unflatten(stagger=10)
        dot.render(output_name, format='png', cleanup=True)
        

    def get_adj_list(self) -> Dict:
        '''
        Returns an adjacency list with all the edges reversed
        '''
        res = {}
        for e in self.edges:
            if e.in_node in res:
                res[e.in_node].append(e.out_node)
            else:
                res[e.in_node] = [e.out_node]
        return res
    

